/*
takepayment.c - Remove HashCash/MIME payment for services rendered
Copyright (C) 2003 Ian Gulliver

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include "firemake.h"
#include "sha1.h"

#define MAX_TOKEN_LEN 256
#define MAXBITS (SHA1_DIGEST_BYTES * 8)
#define bit(c,n) (((c) >> (7 - (n))) & 1)

int gargc;
char **gargv;
struct firestring_estr_t mintbody = {0}, paymentheader = {0}, paymentbody;
struct firestring_estr_t boundary = {0};
struct firestring_estr_t message;
int bits;

int bitcmp(char *a, char *b) {
	int bits = 0;
	while (bits < MAXBITS) {
		if (a[0] != b[0])
			break;
		bits += 8;
		a++;
		b++;
	}
	while (bits < MAXBITS) {
		if (bit(a[0],bits % 8) != bit(b[0],bits % 8))
			break;
		bits++;
	}
	return bits;
}

void foldback(struct firestring_estr_t *headers, struct firestring_estr_t *name) {
	struct firestring_estr_t *fullheader;

	fullheader = firemime_get_full_header(headers,name,&ESTR_S("\n"));
	if (fullheader == NULL)
		return;

	memmove(fullheader->s,fullheader->s + fullheader->l,headers->l - (fullheader->s - headers->s + fullheader->l));
	headers->l -= fullheader->l;
}

int printhandler(struct firemime_part *stack, int depth, void *opaque) {
	if (depth == 0) {
		struct firestring_estr_t *tempvalue;
		firestring_fprintf(stdout,"%e\n",&stack[depth].headers);
		tempvalue = firemime_param_value(stack[0].type_params,&ESTR_S("boundary"));
		if (tempvalue == NULL) {
			firestring_fprintf(stderr,"FireMIME is finding multipart sections without a boundary!\n");
			firestring_estr_free(&message);
			exit(100);
		}
		firestring_estr_aestrcpy(&boundary,tempvalue,0);
		return 0;
	}

	if (paymentheader.s == stack[depth].headers.s) {
		/* this is the header we're concerned about */
		char hostname[256];
		int remove = atoi(gargv[1]);
		char date[32];
		time_t t;
		struct tm *tm;

		if (gethostname(hostname,256) != 0) {
			perror("Unable to get hostname");
			firestring_estr_free(&message);
			exit(100);
		}

		time(&t);
		tm = gmtime(&t);
		strftime(date,32,"%Y%m%dT%H:%M:%S",tm);


		foldback(&stack[depth].headers,&ESTR_S("Amount"));

		firestring_fprintf(stdout,"--%e\nAmount: %d\nAmount-Deducted: %d; hostname=\"%s\"; software=\"takepayment program from FirePay package, http://firestuff.org/\"; reason=\"%s\"; date=\"%s\"\n%e\n%e\n",&boundary,bits - remove,remove,hostname,gargv[2],date,&stack[depth].headers,&stack[depth].body);
		return 1;
	}

	firestring_fprintf(stdout,"--%e\n%e\n%e\n",&boundary,&stack[depth].headers,&stack[depth].body);

	return 1;
}

int handler(struct firemime_part *stack, int depth, void *opaque) {
	if (depth == 0) {
		if (firestring_estr_estrcasecmp(&stack[depth].type,&ESTR_S("multipart"),0) != 0 ||
				firestring_estr_estrcasecmp(&stack[depth].subtype,&ESTR_S("postage"),0) != 0) {
			firestring_fprintf(stderr,"No postage found.\n");
			firestring_estr_free(&message);
			exit(100);
		}
		return 0;
	} else { /* depth is 1 */
		if (firestring_estr_estrcasecmp(&stack[depth].type,&ESTR_S("application"),0) == 0 &&
				firestring_estr_starts(&stack[depth].subtype,"postage-") == 0) {
			/* postage */
			struct firestring_estr_t *version, *to;
			int n;

			if (firestring_estr_estrcasecmp(&stack[depth].subtype,&ESTR_S("postage-hashcash"),0) != 0)
				return 1;

			/* verify version header */
			version = firemime_get_header(&stack[depth].headers,&ESTR_S("Version"),&ESTR_S("\n"));
			if (version == NULL) {
				firestring_fprintf(stderr,"Missing required header field: Version\n");
				firestring_estr_free(&message);
				exit(100);
			}
			if (firestring_estr_estrcmp(version,&ESTR_S("1.0"),0) != 0 && firestring_estr_estrcmp(version,&ESTR_S("1.1"),0) != 0)
				return 1;

			/* check if this is to us */
			to = firemime_get_header(&stack[depth].headers,&ESTR_S("To"),&ESTR_S("\n"));
			if (to == NULL) {
				firestring_fprintf(stderr,"Missing required header field: To\n");
				firestring_estr_free(&message);
				exit(100);
			}
			for (n = 3; n < gargc; n++) {
				if (firestring_estr_strcasecmp(to,gargv[n]) == 0) {
					/* postage to us */
					memcpy(&paymentheader,&stack[depth].headers,sizeof(paymentheader));
					memcpy(&paymentbody,&stack[depth].body,sizeof(paymentbody));
				}
			}
		} else {
			/* mint */
			if (mintbody.s != NULL) {
				firestring_fprintf(stderr,"More than one non-postage object in multipart/postage\n");
				firestring_estr_free(&message);
				exit(100);
			}
			memcpy(&mintbody,&stack[depth].body,sizeof(mintbody));
		}
		return 1;
	}
}

int main(int argc, char *argv[]) {
	struct firestring_estr_t wirebody = {0};
	struct firestring_estr_t fullmint = {0};
	struct firestring_estr_t minthash = {0};
	struct firestring_estr_t *tempheader, date;
	int l,r;

	if (argc < 4) {
		firestring_fprintf(stderr,"Usage: %s <amount> <service> <local address> [<local address> ...]\n",argv[0]);
		exit(100);
	}

	gargc = argc;
	gargv = argv;

	firestring_estr_alloc(&message,4096);
	while ((l = read(0,&message.s[message.l],message.a - message.l)) > 0) {
		message.l += l;
		firestring_estr_expand(&message,message.l + 4096);
	}

	/* we now have a message; call the magic parser */

	if ((r = firemime_parse(&message,handler,&ESTR_S("\n"),NULL)) != 1) {
		firestring_fprintf(stderr,"Error in MIME parsing: %s\n",firemime_strerror(r));
		firestring_estr_free(&message);
		exit(100);
	}

	if (mintbody.s == NULL) {
		firestring_fprintf(stderr,"Missing message body\n");
		firestring_estr_free(&message);
		exit(100);
	}

	if (paymentheader.s == NULL) {
		firestring_fprintf(stderr,"No postage for this destination found\n");
		firestring_estr_free(&message);
		exit(100);
	}

	if (paymentbody.l == 0) {
		firestring_fprintf(stderr,"Empty payment attack detected\n");
		firestring_estr_free(&message);
		exit(100);
	}

	tempheader = firemime_get_header(&paymentheader,&ESTR_S("Date"),&ESTR_S("\n"));
	if (tempheader == NULL) {
		firestring_fprintf(stderr,"Missing required header field: Date\n");
		firestring_estr_free(&message);
		exit(100);
	}
	memcpy(&date,tempheader,sizeof(date));

	tempheader = firemime_get_header(&paymentheader,&ESTR_S("Mint-Hash"),&ESTR_S("\n"));
	if (tempheader != NULL) {
		firestring_estr_alloc(&minthash,tempheader->l * 3 / 4 + 3);
		firestring_estr_base64_decode(&minthash,tempheader);
	}

	firestring_estr_areplace(&wirebody,&mintbody,&ESTR_S("\r\n"),&ESTR_S("\n"));
	firestring_estr_aestrcpy(&fullmint,firemime_get_header(&paymentheader,&ESTR_S("To"),&ESTR_S("\n")),0);
	firestring_estr_aestrcat(&fullmint,&ESTR_S("\r\n"),0);
	firestring_estr_aestrcat(&fullmint,&date,0);
	firestring_estr_aestrcat(&fullmint,&ESTR_S("\r\n"),0);
	firestring_estr_aestrcat(&fullmint,&wirebody,0);
	firestring_estr_free(&wirebody);

	{
		struct tm tm;
		time_t t,n;
		char d[18];

		putenv("TZ=UTC");
		if (date.l != 17) {
			firestring_fprintf(stderr,"Invalid Date header length in postage (%d, '%e')\n",date.l,&date);
			firestring_estr_free(&message);
			exit(100);
		}

		memcpy(d,date.s,17);
		d[17] = '\0';
		/* Date: 20030723T16:04:31
		 *       01234567890123456
		 */

		tm.tm_isdst = 0;
		tm.tm_sec = atoi(&d[15]);
		d[14] = '\0';
		tm.tm_min = atoi(&d[12]);
		d[11] = '\0';
		tm.tm_hour = atoi(&d[9]);
		d[8] = '\0';
		tm.tm_mday = atoi(&d[6]);
		d[6] = '\0';
		tm.tm_mon = atoi(&d[4]) - 1;
		d[4] = '\0';
		tm.tm_year = atoi(d) - 1900;

		t = mktime(&tm);
		if (t == -1) {
			firestring_fprintf(stderr,"Invalid Date header format in postage\n");
			firestring_estr_free(&message);
			exit(100);
		}

		time(&n);
		if (t > n + 3600) {
			firestring_fprintf(stderr,"Postage date is in the future; please check system clock\n");
			firestring_estr_free(&message);
			exit(100);
		}
		if (t < n - 604800) {
			firestring_fprintf(stderr,"Postage date is more than one week in the past; please check system clock or resend message\n");
			firestring_estr_free(&message);
			exit(100);
		}
	}


	/* we now have the full item to mint */
	{
		SHA1_ctx ctx1,ctx2;
		char o1[SHA1_DIGEST_BYTES], o2[SHA1_DIGEST_BYTES];

		SHA1_Init(&ctx1);
		SHA1_Update(&ctx1,fullmint.s,fullmint.l);
		firestring_estr_free(&fullmint);
		memcpy(&ctx2,&ctx1,sizeof(ctx2));
		SHA1_Final(&ctx1,(byte *)o1);
		if (minthash.l > 0) {
			if (minthash.l != SHA1_DIGEST_BYTES) {
				firestring_fprintf(stderr,"Invalid Mint-Hash header contents\n");
				firestring_estr_free(&message);
				exit(100);
			}
			if (memcmp(minthash.s,o1,SHA1_DIGEST_BYTES) != 0) {
				firestring_fprintf(stderr,"Message contents changed in transit; hashcash payment is invalidated\n");
				firestring_estr_free(&message);
				exit(100);
			}
		}


		SHA1_Update(&ctx2,paymentbody.s,paymentbody.l);
		SHA1_Final(&ctx2,(byte *)o2);

		bits = bitcmp(o1,o2);
		{
			struct firestring_estr_t *amount;
			amount = firemime_get_header(&paymentheader,&ESTR_S("Amount"),&ESTR_S("\n"));
			if (amount != NULL) {
				int a;
				amount->a++; /* can never run into another header */
				firestring_estr_0(amount);
				a = atoi(amount->s);
				if (a > bits) {
					firestring_fprintf(stderr,"Attempt to cheat Amount: header higher than actual bits (%d header/%d actual)\n",a,bits);
					firestring_estr_free(&message);
					exit(100);
				}
				bits = a;
				amount->s[amount->a-1] = '\n';
			}
		}
		if (bits < atoi(argv[1])) {
			firestring_fprintf(stderr,"Insufficient postage found (%d needed, %d found)\n",atoi(argv[1]),bits);
			firestring_estr_free(&message);
			exit(100);
		}
	}

	if ((r = firemime_parse(&message,printhandler,&ESTR_S("\n"),NULL)) != 1) {
		firestring_fprintf(stderr,"Error in MIME parsing: %s\n",firemime_strerror(r));
		firestring_estr_free(&message);
		exit(100);
	}

	firestring_fprintf(stdout,"--%e--\n",&boundary);
	firestring_estr_free(&boundary);
	firestring_estr_free(&message);

	return 0;
}
